/* mta-connect:
 *
 * MTA-connect is a generic SMTP conversation library, primarly
 * focused at /sending/ mailpacks to existing server. Currently it
 * is used by Xamime and inflex [via Mailfeeder]
 *
 * Platform Porting Table:
 *----------------------------------------------------
 * Linux: OK
 * FreeBSD: OK ( requires sys/time.h for various calls )
 * Solaris: OK ( requires -lnsl -lsocket and netdb.h header )
 *
 * Started : 14/01/2002
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#if OSTYPE==FreeBSD
#include <sys/time.h>
#endif
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>
#include <string.h>

#include "pldstr.h"
#include "logger.h"
#include "mta-connect.h"


#ifndef FL
#define FL __FILE__,__LINE__
#endif

#define MTACON_DVERBOSE ((MTACON_debug & MTACON_DEBUG_VERBOSE) == MTACON_DEBUG_VERBOSE)
#define MTACON_DNORMAL ((MTACON_debug & MTACON_DEBUG_NORMAL) == MTACON_DEBUG_NORMAL)
#define MTACON_DPEDANTIC ((MTACON_debug & MTACON_DEBUG_PEDANTIC) == MTACON_DEBUG_PEDANTIC)

static int MTACON_debug = 0;
static int MTACON_verbose = 0;
static int MTACON_conversation = 0;

// BEGIN SMP MOD
int lastline_eol = 1;
// END SMP MOD

/*------------------------------------------------------------------------
Procedure:     MTACON_set_debug ID:1
Purpose:       Turns on/off the debugging messages within MTACON.
Input:
Output:
Errors:
------------------------------------------------------------------------*/
int MTACON_set_debug( int level )
{
	MTACON_debug = level;

	return 0;
}


int MTACON_set_conversation( int level )
{
	MTACON_conversation = level;

	return 0;
}

/*------------------------------------------------------------------------
Procedure:     MTACON_get_response ID:1
Purpose:       Used to absorb/accept data being sent from the SMTP server.
Note, this data is ignored, as we are simply following SMTP
protocol and assume that if email has gotten to here that it must
have been accepted by postfix and Xamime already.
Input:
Output:
Errors:
------------------------------------------------------------------------*/
int MTACON_get_response( int pfsocket )
{
	char buf[1024];
	int result;
	struct timeval tv;
	int timeout=30;
	fd_set recv_fdset;

	// Clear our set of selectors
	FD_ZERO(&recv_fdset);

	// Add in our pfsocket fd.
	FD_SET(pfsocket, &recv_fdset);

	// Set our timeout
	tv.tv_sec = 30;
	tv.tv_usec = 0;

	// Infinately loop until either we time out or we get the
	// data we were after.

	while (1)
	{
		// Set up the select to wait for activity on our socket

		result=select(pfsocket+1,&recv_fdset, NULL, NULL, &tv);

		// If we got a non-zero result, that's good!
		if (result == -1)
		{
			LOGGER_log("%s:%d:MTACON_get_response:ERROR: Select on socket returned -1 (%s)",FL,strerror(errno));
			return -1;
		} else if (result == 0) { 
			LOGGER_log("%s:%d:MTACON_get_response:WARNING: No response after %d seconds, aborting attempt", FL, timeout );
			return -1;
		} else {
			int read_size;

			// Read that data from the socket.
			read_size=read(pfsocket, buf, sizeof(buf));

			// Put a string termination on
			buf[read_size] = '\0';

			if (MTACON_conversation) LOGGER_log("<< %s", buf);

			// Send out what we received, for debugging
			if (MTACON_debug) LOGGER_log("%s:%d:MTACON_get_response:DEBUG: Read: %s",FL,buf);

			// If the last char is a \n or \r, break, as we've gotten
			// what we came for.
			if ((buf[read_size-1] == '\n')||(buf[read_size-1] == '\r')) break;
		}
	}
	return result;
}

/*------------------------------------------------------------------------
Procedure:     MTACON_make_call ID:1
Purpose:       Makes a connection to port required for
sending the data to.
Input:
Output:
Errors:
------------------------------------------------------------------------*/
int MTACON_make_call(char *hostname, unsigned short portnum)
{
	struct sockaddr_in sa;
	struct hostent     *hp;
	int s;
	int result;

	if (MTACON_DNORMAL) LOGGER_log("MTACON_make_call: Getting host ID");
	hp = gethostbyname(hostname);

	if (hp == NULL)
	{
		errno= ECONNREFUSED;
		return -1;
	}

	// Clear the struct
	memset(&sa,0,sizeof(sa));

	// Set address.
	if (MTACON_DNORMAL) LOGGER_log("MTACON_make_call: setting connect structure details");
	memcpy((char *)&sa.sin_addr,hp->h_addr,hp->h_length);
	sa.sin_family= hp->h_addrtype;
	sa.sin_port= htons((u_short)portnum);

	// Attempt to make the socket
	if (MTACON_DNORMAL) LOGGER_log("MTACON_make_call: Making socket");
	s = socket(hp->h_addrtype,SOCK_STREAM,0);
	if (s < 0)
	{
		LOGGER_log("%s:%d:MTACON_make_call:ERROR: Unable to get socket to communicate (%s)",FL,strerror(errno));
		return -1;
	}

	// Attempt to connect
	if (MTACON_DNORMAL) LOGGER_log("MTACON_make_call: Attempting connect");
	if (MTACON_conversation) LOGGER_log("Connecting...");

	result = connect(s,(struct sockaddr *)&sa,sizeof(sa));
	if (result < 0)
	{
		close(s); 
		// We only put this log output on the 'verbose' level because we don't want connection
		//	retries spilling out needless reports to stdout or where ever - instead, we'll
		// let the return code indicate that there's a connection problem to the calling parent
		// and let them decide what reports they want to give out
		if (MTACON_DVERBOSE) LOGGER_log("%s:%d:MTACON_make_call:ERROR: Unable to connect (%s)",FL,strerror(errno));
		return -1;
	}

	// If all went well, we return the socket identifier.
	return s;
}




/*------------------------------------------------------------------------
Procedure:     MTACON_send_mailpack ID:1
Purpose:       Sends a given text file/mailpack through the
supplied socket file-descriptor.
Input:
Output:
Errors:
------------------------------------------------------------------------*/
int MTACON_send_mailpack( char *mailpackname, char *sender, char *receiver, char *domain, int sk )
{
	struct PLD_strtok st;
	unsigned char buf[1024];
	int result = 0, wresult = 0;
	int mta_result;
	FILE *fin;

	// BEGIN SMP MOD
	char workbuf[1024];
	char *p;
	char *delims = { " " };
	// END SMP MOD

	if (sk < 0)
	{
		LOGGER_log("%s:%d:MTACON_send_mailpack:ERROR: Socket supplied is invalid (%s)",FL,sk);
		return -1;
	}

	// Open the mailpack we're trying to send to.

	// if our mailpack is coming via stdin, then make *fin point
	// to it
	if (*mailpackname == '-')
	{
		fin = stdin;
	}
	else
	{
		fin = fopen(mailpackname,"r");
		if (!fin)
		{
			LOGGER_log("%s:%d:MTACON_send_mailpack:ERROR: while opening mailpack '%s' for reading. (%s)",FL, mailpackname, strerror(errno));
			return -1;
		}
	}

	// Send the preliminary headers to get things going

	// Get the SMTP introduction message
	mta_result = MTACON_get_response(sk);
	if (mta_result == -1)
	{
		LOGGER_log("%s:%d:MTACON_send_mailpack:ERROR: While waiting on SMTP intro",FL);
		return -1;
	}

	// Introduce ourselves
	snprintf(buf,sizeof(buf),"HELO %s\r\n", domain );
	write(sk, buf, strlen(buf));
	if (MTACON_conversation) LOGGER_log(">> %s", buf);
	mta_result = MTACON_get_response(sk);
	if (mta_result == -1)
	{
		LOGGER_log("%s:%d:MTACON_send_mailpack:ERROR: While on HELO response",FL);
		return -1;
	}

	// Say who we're sending email from...
	snprintf(buf,sizeof(buf),"mail from: %s\r\n", sender);
	write(sk,buf,strlen(buf));
	if (MTACON_conversation) LOGGER_log(">> %s", buf);
	mta_result = MTACON_get_response(sk);
	if (mta_result == -1)
	{
		LOGGER_log("%s:%d:MTACON_send_mailpack:ERROR: While waiting on MAIL FROM response",FL);
		return -1;
	}

	// Say who we're sending it to...
	//
	// BEGIN SMP MOD
	// Modified for multiple recipients
	// 
	p = PLD_strtok( &st, receiver, delims );
	while( p != NULL ) 
	{
		snprintf(buf,sizeof(buf),"rcpt to: %s\r\n", p);
		write(sk,buf,strlen(buf));
		if (MTACON_conversation) LOGGER_log(">> %s", buf);
		mta_result = MTACON_get_response(sk);
	if (mta_result == -1)
	{
		LOGGER_log("%s:%d:MTACON_send_mailpack:ERROR: While waiting on RCPT reponse",FL);
		return -1;
	}
		p = PLD_strtok( &st, NULL, delims );
	}
	// END SMP MOD

	// Tell SMTP server that we're about to send the data
	snprintf(buf,sizeof(buf),"data\r\n");
	write(sk,buf,strlen(buf));
	if (MTACON_conversation) LOGGER_log(">> %s", buf);
	mta_result = MTACON_get_response(sk);
	if (mta_result == -1)
	{
		LOGGER_log("%s:%d:MTACON_send_mailpack:ERROR: While waiting on DATA response",FL);
		return -1;
	}

	// Send the data...
	//	while ((result = fread(buf, 1, 1024, fin)))
	//	BEGIN SMP MOD
	while ( fgets(buf, sizeof(buf), fin))
	{
		// Only use 1023 characters in case we need to prepend
		// a period...
		//
		// RFC 2821, section 4.5.2, requires the sending agent to 
		// examine the first character of a line, and if it is a
		// period, insert an additional period in front of it.
		// This is done even if there is additional text on the
		// line. The SMTP receiver is required to similarly delete
		// the first character of any line that starts with a period.
		//
		// the fgets call is not guaranteed to really end on a line,
		// so I use a flag... On any given read, the flag will be
		// from the last read, so we'll know if the last time through
		// had a end-of-line delimiter. If so, then we can consider
		// this a case needing to be fixed, otherwise, just continue
		// to read as usual...
		//
		if (lastline_eol) {
			if (buf[0] == '.') {
				// save our current line
				strcpy(workbuf,buf);
				// put a period in the outbound line, 
				// then add the original source line.
				strcpy(buf,".");
				strcat(buf,workbuf);
			}
		}
		// Now, after we've evaluated the line, set the flag as to
		// whether or not this line had a CR or LF for next pass...
		//
		// 20030327-0704:PLD:Changed 'index()' to strchr() for reasons
		//		of greater portability, index() is conforming to BSD4.3
		//		where's strchr() conforms to SVID 3, POSIX, BSD4.3 and ISO 9899

		if (strchr(buf,'\r') || strchr(buf,'\n')) {
			lastline_eol = 1;
		} else {
			lastline_eol = 0;
		}
		// set result to be our string length...
		result=strlen(buf);

		// continue as normal...
		// END SMP MOD
		wresult = write(sk, buf, result);
		if (MTACON_conversation) LOGGER_log(">> %s", buf);
		if (wresult != result)
		{
			LOGGER_log("%s:%d:MTACON_send_mailpack:WARNING: tried to send %d bytes, but could only send %d",FL,result,wresult);
		}
	}

	// Send the trailing dot and the exit "quit"
	snprintf(buf,sizeof(buf),"\r\n.\nquit\r\n");
	result = write(sk,buf,strlen(buf));
	if (MTACON_conversation) LOGGER_log(">> %s", buf);
	mta_result = MTACON_get_response(sk);
	if (mta_result == -1)
	{
		LOGGER_log("%s:%d:MTACON_send_mailpack:ERROR: While waiting on QUIT response",FL);
		return -1;
	}

	// Close the mailpack file if required.
	if (*mailpackname == '-')
	{
		fclose(fin);
	}

	return result;
}



/*------------------------------------------------------------------------
Procedure:     MTACON_close ID:1
Purpose:       Close the socket handle handed in the
params. NOTE - this currently is a fairly
trivial call (redundant), but we use it anyhow
in case we do have to do some more complex
things later.
Input:
Output:
Errors:
------------------------------------------------------------------------*/
int MTACON_close( int pfsocket )
{
	if (MTACON_conversation) LOGGER_log("Closing connection.");
	close(pfsocket);

	return 0;
}



//--END.

